'''
Gorgon. A simple multiplier process analysis tool in Python
'''
from threading import Thread
from multiprocessing import Process, Queue
from copy import deepcopy
import sys
if sys.version > '3':
    from gorgon.report import GorgonReport
else:
    from report import GorgonReport
from uuid import uuid4
import inspect
from time import sleep
from random import randint
from collections import namedtuple
import imp

ClusterNode = namedtuple('ClusterNode', ('host', 'user', 'key', 'interpreter'))

QUEUE = Queue()
RAND_DELAY_BASE = 0.001


def run_ops_thread(operation, seed, num_ops, report, delay):
    if delay:
        sleep(delay)

    for i in range(num_ops):
        number = seed + i
        call_id = str(uuid4())
        report.start_call(call_id)
        try:
            result = operation(number)
        except Exception as exc:
            result = exc

        report.end_call(call_id, str(result))


def start_process(operation, seed, num_ops, num_threads, random_delay):
    report = GorgonReport()
    num_ops_thread = num_ops // num_threads
    threads = []
    delay = 0
    for i in range(num_threads):
        thread_seed = seed + i * num_ops_thread
        if random_delay:
            delay = randint(0, num_threads) * RAND_DELAY_BASE
        args = (operation, thread_seed, num_ops_thread, report,
                delay)
        thread = Thread(target=run_ops_thread, args=args)
        thread.start()
        threads.append(thread)

    # Wait for all threads to complete
    for thread in threads:
        thread.join()

    QUEUE.put(report)

    return report


TEMP_FILE_NAME = '/tmp/gorgon_file.py'


def run_cluster():
    # Read from the TEMP file
    # Get the parameters
    exec_file = sys.argv[1]
    fname = sys.argv[2]
    seed = int(sys.argv[3])
    num_ops = int(sys.argv[4])
    num_proc = int(sys.argv[5])
    num_threads = int(sys.argv[6])

    # Get the function from the temp file
    module = imp.load_source('operation', exec_file)
    operation = getattr(module, fname)

    # Initialize a Gorgon module and run the test
    harness = Gorgon(operation)
    harness.go(num_operations=num_ops, num_processes=num_proc,
               num_threads=num_threads, seed=seed, silent=True)
    print(harness.cluster_report())

class Gorgon(object):

    def __init__(self, operation):
        try:
            sig = inspect.signature(operation)
            if len(sig.parameters) != 1:
                raise Exception('A callable with one parameter '
                                'should be passed as operation')

        except TypeError:
            raise Exception('A callable should be passed as operation')
        except AttributeError:
            # This is Python2
            if not inspect.isfunction(operation):
                raise Exception('A callable should be passed as operation')
            if len(inspect.getargspec(operation).args) != 1:
                raise Exception('A callable with one parameter '
                                'should be passed as operation')


        self.operation = operation
        self.processes = []
        self.report = GorgonReport()
        # Store the source code of the call
        self._code = inspect.getsource(operation)
        self._code_name = operation.__name__
        self._cluster = None

    @property
    def num_nodes(self):
        if self._cluster is None:
            return None
        return len(self._cluster)

    def add_to_cluster(self, host, ssh_user, ssh_key_filename,
                       python_interpreter=None):
        try:
            import paramiko
        except ImportError:
            print('paramiko is needed for cluster support')
            raise

        node = ClusterNode(host=host, user=ssh_user,
                           key=ssh_key_filename,
                           interpreter=python_interpreter)
        if not self._cluster:
            self._cluster = []

        self._cluster.append(node)

    def remove_from_cluster(self, host):
        '''
        Remove a node from the cluster. If called with 'all',
        removes all nodes
        '''
        if host == 'all':
            self._cluster = None

        # TODO: Remove from the _cluster array

    def start_pool(self):
        ''' Create a proper pool of workers '''

        for i in range(self.num_processes):
            args = (self.operation,
                    self.seed + (i * self.ops_per_process),
                    self.ops_per_process,
                    self.num_threads,
                    self.random_delay)
            process = Process(target=start_process,
                            args=args)
            process.start()
            self.processes.append(process)

    def insert(self, priority, profile):
        self.queue.put((priority, deepcopy(profile)))

    def wait_until_finish(self):
        # This sometimes fail
        # for process in self.processes:
        #     process.join()

        num_reports = 0
        # Get all the reports on the queue
        while num_reports < self.num_processes:
            report = QUEUE.get()
            self.report.append(report)
            num_reports += 1

    def check_go(self):
        '''
        Check that the number of operations, processes and
        threads adds up
        '''
        cluster_factor = self.num_nodes if self._cluster else 1
        workers = self.num_threads * self.num_processes * cluster_factor
        if self.num_operations % workers != 0:
            close = (self.num_operations // workers) + 1
            suggestion = workers * close
            msg = 'Incorrect number of operations, maybe you mean {}?'
            raise Exception(msg.format(suggestion))

    def go(self, num_operations=1, num_processes=1, num_threads=1,
           seed=0, random_delay=False, silent=False):
        ''' Start the operations '''
        self.num_processes = num_processes
        self.num_threads = num_threads
        self.num_operations = num_operations
        self.seed = seed
        self.check_go()

        if self._cluster is not None:
            self.ops_per_node = num_operations // self.num_nodes
            self.start_cluster()
        else:
            # Start in non clustered way
            self.ops_per_process = num_operations // num_processes
            # Check the number matches
            self.random_delay = random_delay

            if not silent:
                self.print_report_header()

            self.start_pool()

            self.wait_until_finish()

    def start_cluster(self):
        import paramiko
        # Create the temp file and store the code
        tmp_file_name = TEMP_FILE_NAME
        with open(tmp_file_name, 'w') as fp:
            fp.write(self._code)

        # Send the code to all the clusters and start gorgon in cluster mode
        # there
        clients = []
        for node in self._cluster:
            # Connect to remote host
            client = paramiko.SSHClient()
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            client.connect(node.host, username=node.user,
                           key_filename=node.key)
            client.python_interpreter = node.interpreter
            clients.append(client)

        for client in clients:
            # Setup sftp connection and transmit this script
            sftp = client.open_sftp()
            sftp.put(tmp_file_name, tmp_file_name)
            sftp.close()

        stdouts = []
        stderrs = []
        base_seed = self.seed
        for index, client in enumerate(clients):
            run_cmd = 'gorgon_cluster'
            if client.python_interpreter:
                TMPL =  '{} -m gorgon.cluster_run'
                run_cmd =TMPL.format(client.python_interpreter)
            # Run the transmitted script remotely and show its output.
            # SSHClient.exec_command() returns the tuple (stdin,stdout,stderr)
            EXEC_TMPL = ('{run_command} {exec_file} {fname} {seed} {num_ops} '
                         '{num_proc} {num_thread}'
                        )
            seed = base_seed + index * self.ops_per_node
            exec_command = EXEC_TMPL.format(run_command=run_cmd,
                                            exec_file=tmp_file_name,
                                            fname=self._code_name,
                                            seed=seed,
                                            num_ops=self.ops_per_node,
                                            num_proc=self.num_processes,
                                            num_thread=self.num_threads)

            stdin, stdout, stderr = client.exec_command(exec_command)
            stdouts.append(stdout)
            stderrs.append(stderr)

        error = False
        for stderr in stderrs:
            for line in stderr:
                error = True
                print(line)

        if error:
            msg = ('Error while executing remote commands, please check '
                   'commands')
            raise Exception(msg)

        # Obtain the results through stdout
        for stdout in stdouts:
            for line in stdout:
                self.report.append_cluster(line)

        # cleanup
        for client in clients:
            client.exec_command('rm {}*'.format(tmp_file_name))
            client.close()

    def small_report(self):
        return self.report.small_report()

    def html_graph_report(self):
        return self.report.html_graph_report()

    def cluster_report(self):
        return self.report.cluster_report()

    def print_report_header(self):
        TMPL = 'Run operation {} times, with {} processes * {} threads'
        msg = TMPL.format(self.num_operations,
                          self.num_processes,
                          self.num_threads)
        print(msg)

    def http_call(self, params):
        '''
        This function abstracts the access to http calls
        '''
        call_id = str(uuid4())
        # TODO: If url is not present, log an error
        url = params.pop('url')
        method = params.pop('method')
        requests_function = self.MAP_METHODS[method]
        self.report.start_call(call_id, url)
        result = requests_function(url, **params)
        # result = 200
        self.report.end_call(call_id)
        result.previous_url = url
        result.previous_method = method
        return result



import json
from time import time
from collections import deque, defaultdict

HTML_TEMPLATE = '''
<html>
  <head>
    <script type="text/javascript" src="https://www.google.com/jsapi"></script>
    <script type="text/javascript">
      google.load("visualization", "1", {{packages:["corechart"]}});
      google.setOnLoadCallback(drawChart);
      function drawChart() {{
        var data = google.visualization.arrayToDataTable([
          [{titles}],
          {data}
        ]);

        var options = {{
            title: 'Gorgon',
            //vAxes:[{{}}, {{}}],
            //series: [{{targetAxisIndex:0}}, {{targetAxisIndex:1}}]
        }};

        var chart = new google.visualization.LineChart(document.getElementById('chart_div'));
        chart.draw(data, options);
      }}
    </script>
  </head>
  <body>
    <div id="chart_div" style="width: 1500px; height: 800px;"></div>
  </body>
</html>

'''

def average(data):
    total_sum = 0
    total_len = 0
    for d in data:
        total_sum += d
        total_len += 1

    return total_sum // total_len


class GorgonReport(object):

    def __init__(self):
        self.start_time = None
        self.end_time = None
        self.calls = deque()
        self.full_report = []

    def end(self):
        # Create the report
        self.create_report()

    @property
    def total_time(self):
        return self.end_time - self.start_time

    @property
    def formatted_start_time(self):
        return self.start_time

    @property
    def formatted_end_time(self):
        return self.end_time

    TINY_FORMAT = '{ms: 3,}ms'
    SMALL_FORMAT = '{seconds: 2,}s {ms: 3,}ms'
    MEDIUM_FORMAT = '{minutes: 2,}m {seconds: 2,}s {ms: 3,}ms'
    BIG_FORMAT = '{hours}h {minutes: 2,}m {seconds: 2,}s {ms: 3,}ms'

    def append(self, report):
        ''' Append a report to this one '''
        self.calls.extend(report.calls)

    def append_cluster(self, report_line):
        report = json.loads(report_line)
        self.calls.extend(report)

    @property
    def formatted_total_time(self):
        return self.formatted_time(self.total_time)

    def formatted_time(self, unformatted_time):
        ms = int((unformatted_time % 1) * 1000)
        total_sec = int(unformatted_time)
        total_minutes = (total_sec // 60)
        hours = (total_minutes // 60)
        seconds = total_sec % 60
        minutes = total_minutes % 60
        if hours and minutes:
            template = self.BIG_FORMAT
        elif minutes:
            template = self.MEDIUM_FORMAT
        elif seconds:
            template = self.SMALL_FORMAT
        else:
            template = self.TINY_FORMAT

        formatted_msg = template.format(hours=hours,
                                        minutes=minutes,
                                        seconds=seconds,
                                        ms=ms)

        return formatted_msg

    def times_report(self, times, start_time, end_time):
        ''' Get a list of times and return a string with the report '''

        number = len(times)
        average = sum(times) / number
        maximum = max(times)
        minimum = min(times)

        if average == 0:
            TEMPLATE = ('{number: 10}  Inf! ops/sec. Avg time: {avg} '
                        'Max: {maximum} Min: {minimum}')
            opssec = None
        else:
            TEMPLATE = ('{number: 10}  {opssec: 7,} ops/sec. '
                        'Avg time: {avg} '
                        'Max: {maximum} Min: {minimum}')
            total_time = end_time - start_time
            opssec = int(number / total_time)

        report = TEMPLATE.format(number=number,
                                 opssec=opssec,
                                 avg=self.formatted_time(average),
                                 maximum=self.formatted_time(maximum),
                                 minimum=self.formatted_time(minimum))
        return report

    def _get_id_calls_start_end(self):
        id_calls = defaultdict(dict)
        start_time = self.calls[0]['start_time']
        end_time = 0
        for item in self.calls:
            id_calls[str(item['call_id'])].update(item)
            if 'start_time' in item:
                start_time = min(start_time, item['start_time'])
            if 'end_time' in item:
                end_time = max(start_time, item['end_time'])

        for item in id_calls.values():
            item['total_time'] = item['end_time'] - item['start_time']

        return id_calls, start_time, end_time

    def _group_by(self, id_calls, group):
        group_calls = defaultdict(list)

        for call_id, item in id_calls.items():
            group_calls[item[group]].append(item)

        return group_calls

    def small_report(self):
        self.full_report = []

        RESULT_HEADER = 'Result'
        len_result = len(RESULT_HEADER)
        id_calls, start_time, end_time = self._get_id_calls_start_end()
        self.start_time = start_time
        self.end_time = end_time
        for call_id, item in id_calls.items():
            len_result = max(len_result, len(str(item['result'])))

        REPORT_TEMPLATE = '{{name:>{}}} {{report}}'.format(len_result)

        group_calls = defaultdict(list)
        total_times = [item['total_time'] for item in id_calls.values()]
        group_calls = self._group_by(id_calls, 'result')

        header = REPORT_TEMPLATE.format(
            name='Result',
            report=self.times_report(total_times,
                                     self.start_time,
                                     self.end_time)
        )
        self.full_report.append(header)

        for group, calls in group_calls.items():
            start_time = min(item['start_time'] for item in calls)
            end_time = max(item['end_time'] for item in calls)
            times = [item['total_time'] for item in calls]

            row = (REPORT_TEMPLATE.format(name=group,
                                          report=self.times_report(times,
                                                                   start_time,
                                                                   end_time)))
            self.full_report.append(row)

        self.num_operations = len(id_calls)

        # Not sure if we should show the start/end time
        # Maybe useful in long tests
        # print('Start time', self.formatted_start_time)
        # print('End time', self.formatted_end_time)

        result = ['Total time: {}'.format(self.formatted_total_time)]
        result.extend(self.full_report)
        return '\n'.join(result)

    def start_call(self, uuid):
        info = {
            'call_id': uuid,
            'start_time': time(),
        }
        self.calls.append(info)

    def end_call(self, uuid, result):
        info = {
            'call_id': uuid,
            'end_time': time(),
            'result': result
        }
        self.calls.append(info)

    def html_graph_report(self):
        '''
        Generate data to be printed as a graph
        '''
        # Group the data by second
        id_calls, start_time, end_time = self._get_id_calls_start_end()
        # First by result
        group_calls = self._group_by(id_calls, 'result')

        # Group end calls in the same second
        by_second = defaultdict(lambda : defaultdict(list))
        for title, results in group_calls.items():
            for result in results:
                timestamp = int(result['end_time'])
                by_second[title][timestamp].append(result)

        # Get info per second
        by_second_info = defaultdict(lambda : defaultdict(list))
        for title, times in by_second.items():
            for timestamp, data in times.items():
                # in ms
                this_second_times = [int(d['total_time'] * 1000)
                                     for d in data]
                info = {
                    'operations': len(data),
                    'average': average(this_second_times),
                    'maximum': max(this_second_times),
                    'minimum': min(this_second_times),
                }
                by_second_info[timestamp][title] = info

        titles = list(by_second.keys())

        formatted_data = []
        for timestamp, tinfo in sorted(by_second_info.items(),
                                       key=lambda x:x[0]):
            data = [int(timestamp - start_time)]
            formatted_ops = [tinfo.get(title, {}).get('operations', 0)
                             for title in titles]
            data += formatted_ops
            formatted_data.append(data)

        page = self.google_chart(titles, formatted_data)
        return page

    def cluster_report(self):
        ''' Return the calls in JSON format to be transmitted '''
        id_calls, start_time, end_time = self._get_id_calls_start_end()
        return json.dumps([v for v in id_calls.values()])

    def google_chart(self, titles, data):
        all_titles = ['Timestamp'] + titles
        TITLE_TMPL = ', '.join('"{}"'.format(t.replace('"', "'"))
                                             for t in all_titles)
        LINE_TMPL = ', '.join('{}' for t in all_titles)
        LINE_FORMAT = "[" + LINE_TMPL + "]"
        format_data = []
        for line in data:
            format_data.append(LINE_FORMAT.format(*line))

        # Remove the last comma
        format_data[-1] = format_data[-1].rstrip(',')

        page = HTML_TEMPLATE.format(data=',\n          '.join(format_data),
                                    titles=TITLE_TMPL)

        return page



from setuptools import setup

setup(
    # metadata
    name='gorgon',
    description='A simple multiplier task analysis tool in Python',
    license='MIT',
    version='0.3.3',
    author='Jaime Buelta',
    author_email='jaime.buelta@gmail.com',
    url='https://github.com/jaimebuelta/gorgon',
    download_url='https://github.com/jaimebuelta/gorgon/tarball/0.3.3',
    platforms='Cross Platform',
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    keywords=['multiprocess', 'multithread', 'analysis'],
    scripts=['scripts/gorgon_cluster'],
    install_requires=['paramiko'],
    packages=['gorgon'],
)
